unit Config;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Spin, Tags, GradPnl, ComCtrls, jpeg;

type
  TConfigForm = class(TForm)
    Label2: TLabel;
    Button1: TButton;
    Button2: TButton;
    Label3: TLabel;
    HTMLCb: TCheckBox;
    Panel1: TPanel;
    Label6: TLabel;
    BGCb: TColorBox;
    TextCb: TColorBox;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    OpenDialog: TOpenDialog;
    Edit1: TEdit;
    IntervalUd: TUpDown;
    Icon32: TImage;
    Edit2: TEdit;
    ImgWidthUd: TUpDown;
    Edit3: TEdit;
    ImgHeightUd: TUpDown;
    TagCb: TCheckBox;
    Label5: TLabel;
    TempDirEdt: TEdit;
    procedure BrowseFileBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    procedure TimerTimer(Sender: Tobject);
  end;

var
  ConfigForm: TConfigForm;

  Timer: TTimer;
  DefaultURL: TFileName;
  HTMLPreview: Boolean;
  TempDir: string;
  BgColor: TColor;
  TextColor: TColor;
  ImgWidth: Integer;
  ImgHeight: Integer;
  ShowTag: Boolean;

  HWindow: HWND;

  CurrentFileName: string;

  AAudioInfo: TAudioInfo;
  FileTagv2: TID3v2;
  AWMAFile: TWMAfile;

implementation

{$R *.dfm}

procedure SetBrowserURL(URL: string);
begin
  SendMessage(HWindow, WM_USER, Integer(PChar(URL)), 241);
end;

function GetFileName: string;
begin
  Result := StrPas(PChar(SendMessage(HWindow, WM_User, SendMessage(HWindow, WM_USER, 0, 125), 211)));
end;

procedure GetFileList(Path: string; var List: TStringList);
var
  Sr: TSearchRec;
begin
  if List <> nil then
  begin
    List.Clear;
    if (FindFirst(Path, faAnyFile, Sr) = 0) then
    begin
      List.Add(Sr.Name);
      while (FindNext(Sr) = 0) do
        List.Add(Sr.Name);
      FindClose(Sr);
    end;
  end;
end;

function ChoosePicture(CurrentPath: string; ThumbInPriority: Boolean = True): string;
const
  FrontStr: array[1..5] of string = ('folder.jpg', 'thumb', 'front', 'avant', 'facade');
var
  I, J: Integer;
  Sr: TSearchRec;
  JpegList: TStringList;
  Found: Boolean;
begin
  if (FindFirst(CurrentPath + {'\} '*.jpg', faAnyFile, Sr) = 0) then
    if FindNext(Sr) = 0 then
    begin
      JpegList := TStringList.Create;
      try
        GetFileList(CurrentPath + {'\} '*.jpg', JpegList);
        Found := False;
        if ThumbInPriority then
          J := Low(FrontStr)
        else
          J := Low(FrontStr) + 1;
        while (J <= High(FrontStr)) and not Found do
        begin
          I := 0;
          while (I < JpegList.Count) and not Found do
          begin
            if Pos(FrontStr[J], LowerCase(JpegList[I])) > 0 then
            begin
              Result := CurrentPath + {'\' +} JpegList[I];
              Found := True;
            end;
            Inc(I);
          end;
          Inc(J);
        end;
        if not Found then
          Result := CurrentPath + {'\' +} JpegList[0];
      finally
        JpegList.Free;
      end;
    end
    else
      Result := CurrentPath + {'\' +} Sr.Name
  else
    Result := '';
end;

function ByteToHexaStr(Value: Byte): string;
const
  Digits: array[0..15] of Char = ('0', '1', '2', '3', '4', '5', '6', '7', '8',
    '9', 'A', 'B', 'C', 'D', 'E', 'F');
begin
  Result := Digits[Value div 16] + Digits[Value mod 16];
end;

function ColorToHTML(Color: TColor): string;
var
  R, G, B: string;
begin
  R := ByteToHexaStr(GetRValue(ColorToRGB(Color)));
  G := ByteToHexaStr(GetGValue(ColorToRGB(Color)));
  B := ByteToHexaStr(GetBValue(ColorToRGB(Color)));
  Result := '#' + R + G + B;
end;

procedure GetFileInfos(Filename: TFilename; var Lines: TStringList);
begin
  Lines.Add(Format('<table width=%d style="FONT-SIZE: 10px; LINE-HEIGHT: 10px">', [ImgWidth]));
  if UpperCase(ExtractFileExt(Filename)) = '.MP3' then
  begin
    if FileTagv2.ReadFromFile(FileName) then
      if FileTagv2.Exists then
      begin
        if FileTagv2.Artist <> '' then
          Lines.Add('<tr><td valign=top>Artist </td><td valign=top>:</td><td> ' + FileTagv2.Artist + '</td></tr>');
        if FileTagv2.Album <> '' then
          Lines.Add('<tr><td valign=top>Album </td><td valign=top>:</td><td> ' + FileTagv2.Album + '</td></tr>');
        if FileTagv2.Title <> '' then
          Lines.Add('<tr><td valign=top>Title </td><td valign=top>:</td><td> ' + FileTagv2.Title + '</td></tr>');
        if FileTagv2.Year <> '' then
          Lines.Add('<tr><td valign=top>Year </td><td valign=top>:</td><td> ' + FileTagv2.Year + '</td></tr>');
        AAudioInfo.LoadFromFile(Filename);
        Lines.Add('<tr><td valign=top width=50>Duration </td><td valign=top width=5>:</td><td> ' + FormatDateTime('nn:ss', (AAudioInfo.MpegDuration div 1000) / (24.0 * 60 * 60)) + '</td></tr>');
      end
      else
      begin
        AAudioInfo.LoadFromFile(Filename);
        if AAudioInfo.ID3.Ok then
        begin
          if AAudioInfo.ID3.Artist <> '' then
            Lines.Add('<tr><td valign=top>Artist </td><td valign=top>:</td><td> ' + AAudioInfo.ID3.Artist + '</td></tr>');
          if AAudioInfo.ID3.Album <> '' then
            Lines.Add('<tr><td valign=top>Album </td><td valign=top>:</td><td> ' + AAudioInfo.ID3.Album + '</td></tr>');
          if AAudioInfo.ID3.Title <> '' then
            Lines.Add('<tr><td valign=top>Title </td><td valign=top>:</td><td> ' + AAudioInfo.ID3.Title + '</td></tr>');
          if AAudioInfo.ID3.Year <> '' then
            Lines.Add('<tr><td valign=top>Year </td><td valign=top>:</td><td> ' + AAudioInfo.ID3.Year + '</td></tr>');
          Lines.Add('<tr><td valign=top width=50>Duration </td><td valign=top width=5>:</td><td> ' + FormatDateTime('nn:ss', (AAudioInfo.MpegDuration div 1000) / (24.0 * 60 * 60)) + '</td></tr>');
        end
        else
        begin
          Lines.Add('<tr><td valign=top>Filename </td><td valign=top>:</td><td> ' + Copy(ExtractFileName(Filename), 1, System.Length(ExtractFileName(Filename)) - System.Length(ExtractFileExt(Filename))) + '</td></tr>');
          Lines.Add('<tr><td valign=top width=50>Duration </td><td valign=top width=5>:</td><td> ' + FormatDateTime('nn:ss', (AAudioInfo.MpegDuration div 1000) / (24.0 * 60 * 60)) + '</td></tr>');
        end;
      end
  end
  else if UpperCase(ExtractFileExt(Filename)) = '.WMA' then
    if AWMAFile.ReadFromFile(FileName) then
      if AWMAFile.Valid then
      begin
        if AWMAFile.Artist <> '' then
          Lines.Add('<tr><td valign=top>Artist </td><td valign=top>:</td><td> ' + AWMAFile.Artist + '</td></tr>');
        if AWMAFile.Album <> '' then
          Lines.Add('<tr><td valign=top>Album </td><td valign=top>:</td><td> ' + AWMAFile.Album + '</td></tr>');
        if AWMAFile.Title <> '' then
          Lines.Add('<tr><td valign=top>Title </td><td valign=top>:</td><td> ' + AWMAFile.Title + '</td></tr>');
        if AWMAFile.Year <> '' then
          Lines.Add('<tr><td valign=top>Year </td><td valign=top>:</td><td> ' + AWMAFile.Year + '</td></tr>');
        Lines.Add('<tr><td valign=top width=50>Duration </td><td valign=top width=5>:</td><td> ' + FormatDateTime('nn:ss', AWMAFile.Duration / (24.0 * 60 * 60)) + '</td></tr>');
      end;
  Lines.Add('</table>');
end;

procedure TConfigForm.TimerTimer(Sender: TObject);
var
  NewFileName: string;
  PicFileStr: string;
  Lines: TStringList;
  Refresh: Boolean;
begin
  Refresh := False;
  if (SendMessage(HWindow, WM_USER, 0, 104) <> 0) then
  begin
    NewFileName := GetFileName;
    if (NewFileName <> CurrentFileName) and
      (ShowTag or (ExtractFilePath(NewFileName) <> ExtractFilePath(CurrentFileName))) then
    begin
      CurrentFileName := NewFileName;
      PicFileStr := ChoosePicture(ExtractFilePath(CurrentFileName));
      if (PicFileStr = '') then
        PicFileStr := DefaultURL;
      Refresh := True;
    end
  end
  else if CurrentFileName <> '' then
  begin
    CurrentFileName := '';
    PicFileStr := DefaultURL;
    Refresh := True;
  end;

  if Refresh then
    if HTMLPreview then
    begin
      Lines := TStringList.Create;
      with Lines do
      try
        Add('<html><head><title> Preview </title></head>');
        Add('<body bgcolor=' + ColorToHTML(BgColor) + ' text=' + ColorToHTML(TextColor) + ' style="PADDING-RIGHT: 0px; PADDING-LEFT: 0px; PADDING-BOTTOM: 0px; MARGIN: 10px 0px; PADDING-TOP: 0px; FONT-FAMILY: verdana, arial, sans-serif;">');
        Add('<div align=center><img width=' + IntToStr(ImgWidth) + ' height=' + IntToStr(ImgHeight) + ' src="' + PicFileStr + '">');
        if ShowTag then
          GetFileInfos(CurrentFileName, Lines);
        Add('</div></body></html>');
        SaveToFile(TempDir + '\coverpreview.html');
        SetBrowserURL('file:///' + TempDir + '\coverpreview.html');
      finally
        Free;
      end;
    end
    else
      SetBrowserURL('file:///' + PicFileStr)
end;

procedure TConfigForm.BrowseFileBtnClick(Sender: TObject);
begin
  {  if OpenDialog.Execute then
      DefaultURLEdt.Text := OpenDialog.FileName;}
end;

procedure TConfigForm.FormCreate(Sender: TObject);
begin
  with TGradientPanel.Create(Self) do
  begin
    Parent := Self;
    Align := alTop;
    Caption := 'HTML Album Cover';
    Font.Name := 'Arial';
    Font.Size := 14;
    Font.Style := [fsBold];
    Font.Color := clWhite;
    FEndColor := clBtnFace;
    ;
    FBeginColor := clBlack;
    Icon := Icon32.Picture.Icon;
  end;
end;

end.

