unit Defs;

interface

uses
  Windows;

function Init: Integer; cdecl;
procedure Config; cdecl;
procedure Quit; cdecl;

type
  PWinampGeneralPurposePlugin = ^TWinampGeneralPurposePlugin;
  TWinampGeneralPurposePlugin = record
    version: Integer;
    description: PChar;
    init: function: Integer; cdecl;
    config: procedure; cdecl;
    quit: procedure; cdecl;
    hwndParent: HWND;
    hDllInstance: HINST;
  end;

const
  WinampGeneralPurposePlugin: TWinampGeneralPurposePlugin =
  (
    version: $10;
    description: 'HTML Album cover (GEN_HTMLCOVER.DLL)';
    init: Init;
    config: Config;
    quit: Quit;
    hwndParent: 0;
    hDllInstance: 0;
    );

function winampGetGeneralPurposePlugin: PWinampGeneralPurposePlugin; cdecl; export;

implementation

uses Messages, Controls, Dialogs, ExtCtrls, Graphics, IniFiles, SysUtils, Config, Tags,
  ComCtrls, StdCtrls;

function Init: Integer; cdecl;
begin
  HWindow := FindWindow(PChar('Winamp v1.x'), nil);
  FileTagv2 := TID3v2.Create;
  AAudioInfo := TAudioInfo.Create;
  AWMAFile := TWMAFile.Create;
  ConfigForm := TConfigForm.Create(nil);
  Timer := TTimer.Create(nil);
  Timer.Enabled := True;
  Timer.OnTimer := ConfigForm.TimerTimer;
  DefaultURL := ExtractFilePath(ParamStr(0)) + 'plugins\gen_HTMLCover_Default.jpg';
  with TIniFile.Create(ExtractFilePath(ParamStr(0)) + 'Winamp.ini') do
  try
    Timer.Interval := ReadInteger('HTMLCover', 'Timer.Interval', 300);
    HTMLPreview := ReadBool('HTMLCover', 'HTMLPreview', False);
    TempDir := ReadString('HTMLCover', 'TempDir', 'C:\WINDOWS\Temp');
    BgColor := StringToColor(ReadString('HTMLCover', 'BgColor', 'clBlack'));
    TextColor := StringToColor(ReadString('HTMLCover', 'TextColor', 'clWhite'));
    ImgWidth := ReadInteger('HTMLCover', 'ImgWidth', 200);
    ImgHeight := ReadInteger('HTMLCover', 'ImgHeight', 200);
    ShowTag := ReadBool('HTMLCover', 'ShowTag', True);
  finally
    Free;
  end;
  CurrentFileName := 'no-file-selected';
  SendMessage(HWindow, WM_USER, Integer(PChar('file:///' + DefaultURL)), 241);
  Result := 0; //the plugin was succesfully initialized
end;

procedure Config; cdecl;
begin
  ConfigForm.IntervalUd.Position := Timer.Interval;
  ConfigForm.HTMLCb.Checked := HTMLPreview;
  ConfigForm.TempDirEdt.Text := TempDir;
  ConfigForm.BGCb.Selected := BgColor;
  ConfigForm.TextCb.Selected := TextColor;
  ConfigForm.ImgWidthUd.Position := ImgWidth;
  ConfigForm.ImgHeightUd.Position := ImgHeight;
  ConfigForm.TagCb.Checked := ShowTag;
  if ConfigForm.ShowModal = mrOk then
  begin
    Timer.Interval := ConfigForm.IntervalUd.Position;
    HTMLPreview := ConfigForm.HTMLCb.Checked;
    TempDir := ConfigForm.TempDirEdt.Text;
    BgColor := ConfigForm.BGCb.Selected;
    TextColor := ConfigForm.TextCb.Selected;
    ImgWidth := ConfigForm.ImgWidthUd.Position;
    ImgHeight := ConfigForm.ImgHeightUd.Position;
    ShowTag := ConfigForm.TagCb.Checked;
    with TIniFile.Create(ExtractFilePath(ParamStr(0)) + 'Winamp.ini') do
    try
      WriteInteger('HTMLCover', 'Timer.Interval', Timer.Interval);
      WriteBool('HTMLCover', 'HTMLPreview', HTMLPreview);
      WriteString('HTMLCover', 'TempDir', TempDir);
      WriteString('HTMLCover', 'BgColor', ColorToString(BgColor));
      WriteString('HTMLCover', 'TextColor', ColorToString(TextColor));
      WriteInteger('HTMLCover', 'ImgWidth', ImgWidth);
      WriteInteger('HTMLCover', 'ImgHeight', ImgHeight);
      WriteBool('HTMLCover', 'ShowTag', ShowTag);
    finally
      Free;
    end;
  end;
end;

procedure Quit; cdecl;
begin
  FileTagv2.Free;
  AAudioInfo.Free;
  AWMAFile.Free;
  Timer.Free;
  ConfigForm.Free;
end;

function winampGetGeneralPurposePlugin: PWinampGeneralPurposePlugin; cdecl;
begin
  Result := @WinampGeneralPurposePlugin;
end;

end.

