{$B-}
{build from :
 - MpegFrameHdr & ID3Tags : Andrey V.Sorokin (anso@mail.ru, anso@usa.net, http://anso.da.ru, http://anso.virtualave.net)
 - ID3v2 & WMAFile : Jurgen Faul (jfaul@gmx.de, http://jfaul.de/atl)}
unit Tags;

interface

uses
 Classes, SysUtils;

const
 MpegFrameHdrSz = 4;

 MPEG_VER_ERROR = 0;
 MPEG_VER_1 = 1;
 MPEG_VER_2 = 2;
 MPEG_VER_2_5 = 3;

 MpegVerStr : array [0 .. 3] of string [3] = (
  '?!!', '1.0', '2.0', '2.5');

 MPEG_LAYER_ERROR = 0;
 MPEG_LAYER_I = 1;
 MPEG_LAYER_II = 2;
 MPEG_LAYER_III = 3;

 MpegLayerStr : array [0 .. 3] of string [3] = (
  '?!!', 'I', 'II', 'III');

 MPEG_BITRATE_ERROR = -1;
 MPEG_SAMPLINGRATE_ERROR = -1;

 MPEG_CHANNEL_STEREO = 0;
 MPEG_CHANNEL_JOINTSTEREO = 1;
 MPEG_CHANNEL_DUAL = 2;
 MPEG_CHANNEL_SINGLE = 3;

 MpegChannelStr : array [0 .. 3] of string = (
  'Stereo', 'Joint Stereo', 'Dual', 'Mono');

 MPEG_EMPHASIS_NONE = 0;
 MPEG_EMPHASIS_50_15 = 1;
 MPEG_EMPHASIS_ERROR = 2;
 MPEG_EMPHASIS_CCIT_J_17 = 3;

 MpegEmphasisStr : array [0 .. 3] of string = (
  'None', '50/15 ms', '?!!', 'CCIT J.17');

  TAG_VERSION_2_3 = 3;                               { Code for ID3v2.3.x tag }

  { Channel modes }
  WMA_CM_UNKNOWN = 0;                                               { Unknown }
  WMA_CM_MONO = 1;                                                     { Mono }
  WMA_CM_STEREO = 2;                                                 { Stereo }

  { Channel mode names }
  WMA_MODE: array [0..2] of string = ('Unknown', 'Mono', 'Stereo');

type

 TMpegFrameHdr = class

   private
    fHdrBuf : array [0 .. MPegFrameHdrSz - 1] of byte;

    // some precalculated fieds for optimization
    // (calculated in Load and Locate)
    fOk : boolean;
    fVersion : integer;
    fKBitRate : integer;
    fSamplingRate : integer;
    fFrameLen : integer;

    // calculating fields "on the fly" from fHdrBuf
    function GetOk : boolean;
    function GetSync : boolean;
    function GetVersion : integer;
    function GetVersionStr : string;
    function GetLayer : integer;
    function GetLayerStr : string;
    function GetCRCExists : boolean;
    function GetKBitRate : integer;
    function GetBitRate : integer;
    function GetSamplingRate : integer;
    function GetPadding : boolean;
    function GetPaddingLen : integer;
    function GetChannelMode : integer;
    function GetChannelModeStr : string;
    function GetChannelModeExt : integer;
    function GetChannelModeExtStr : string;
    function GetCopyright : boolean;
    function GetOriginal : boolean;
    function GetEmphasis : integer;
    function GetEmphasisStr : string;
    function GetFrameLen : integer;

   public

    constructor Create;
    // Simple set Ok to False

    function Load (const ABuf; ABufSz : integer) : boolean;
    // Initialize class fieds with values from
    // mpeg-header AHeader (MPegFrameHdrBufSz bytes)
    // Return true if all fields is Ok.

    function Locate (const ABuf; ABufSz : integer) : integer;
    // Search Mpeg header in ABuf buffer of ABufSz length.
    // If success then call Load and return index of first
    // header byte in APos (0 - index of first byte of ABuf)
    // otherwise return -1

    property Ok : boolean read fOk write fOk;
    // True, if all fealds of header is valid
    // Writeable only for reinitialization purpose.

    property Sync : boolean read GetSync;
    // True, if there is valid sinc bits in header

    property Version : integer read fVersion;
    // MPEG_VER_* value

    property VersionStr : string read GetVersionStr;
    // Version as string ('1.0', '2.0', '2.5', '?!!')

    property Layer : integer read GetLayer;
    // MPEG_LAYER_* value

    property LayerStr : string read GetLayerStr;
    // Layer as string ('I', 'II', 'III', '?!!')

    property CRCExists : boolean read GetCRCExists;
    // True if there is CRC

    property KBitRate : integer read fKBitRate;
    // BitRate in Bits/1000/sec

    property BitRate : integer read GetBitRate;
    // BitRate in Bits/sec

    property SamplingRate : integer read fSamplingRate;
    // Sampling frequency, Hz

    property Padding : boolean read GetPadding;
    // True if there is padding slot in the frame

    property PaddingLen : integer read GetPaddingLen;
    // Length of padding, bytes (zero if no padding)

    property ChannelMode : integer read GetChannelMode;
    // MPEG_CHANNEL_*

    property ChannelModeStr : string read GetChannelModeStr;
    // ChannelMode as english strings (from MpegChannelStr)

    property ChannelModeExt : integer read GetChannelModeExt;
    // Channel mode extension (only for joint stereo)

    property ChannelModeExtStr : string read GetChannelModeExtStr;
    // Channel mode extention as english string

    property Copyright : boolean read GetCopyright;
    // True if copyright

    property Original : boolean read GetOriginal;
    // True if original media

    property Emphasis : integer read GetEmphasis;
    // MPEG_EMPHASIS_*

    property EmphasisStr : string read GetEmphasisStr;
    // Emphasis as english string

    property FrameLen : integer read fFrameLen;
    // length of the frame (bytes)
  end;

  TID3v1Tag = class // ID3v1 & ID3v1.1

   private
    fOk : boolean;
    fModified : boolean;
    fv1 : boolean;
    fTitle : string; // 30
    fArtist : string; // 30
    fAlbum : string; // 30
    fTrack : integer;
    fComment : string; // 30 in ID3v1, 28 in ID3v1.1
    fYear : string; // 4
    fGenreID : integer;
    fGenre : string;

    procedure Setv1 (b : boolean);
    procedure SetField (AIdx : integer; const s : string);
    procedure SetGenreID (AGenreID : integer);
    procedure SetTrack (ATrack : integer);

   public

    constructor Create;

    procedure Load (const ABuf; ABufSz : integer);
    // Load tag from ABuf (max ABufSz)

    procedure LoadFromStream (AStream : TStream; AStreamSz : integer);
    // find ID3tag in AStream (max AStreamSz bytes from curren position)
    // and Load it. If success then returns offset to tag, else -1

    procedure LoadFromFile (const AFileName : string);
    // Load ID3Tag from file AFileName

    procedure Save (var ABuf; ABufSz : integer);
    // Write tag into ABuf (max ABufSz bytes), clear Modified

    procedure SaveToStream (AStream : TStream);
    // Write tag ito current position of AStream, clear Modified

    procedure SaveToFile (const AFileName : string);
    // Add/Update tag in file AFileName, clear Modified

    procedure DeleteFromFile (const AFileName : string);
    // Delete tag from file AFileName (if exists)

    property Ok : boolean read fOk write fOk;
    // True if loaded tag seems to be Ok

    property Modified : boolean read fModified;
    // true if anything changed

    property v1 : boolean read fv1 write Setv1;
    // True if ID3v1, false if ID3v1.1

    property Title : string index 1 read fTitle write SetField;
    // Music title

    property Artist : string index 2 read fArtist write SetField;
    // Performer

    property Year : string index 3 read fYear write SetField;
    // year as 4-chars string

    property Album : string index 4 read fAlbum write SetField;
    // Album title

    property Track : integer read fTrack write SetTrack; // ID3v1.1 only
    // Track number (only if not v1)

    property Comment : string index 5 read fComment write SetField;
    // Comment

    property GenreID : integer read fGenreID write SetGenreID;
    // Genre ID (see ID3Genres withc contains genres
    // list as TStrings)

    property Genre : string read fGenre;
    // Genre string
  end;

  TAudioInfo = class
   private
    fMpg : TMpegFrameHdr;
    fFirstMpegFramePos : integer;
    fXingVBR : boolean;
    fID3 : TID3v1Tag;

    fMpegDuration : integer;

    fMaxGapBeforeMpeg : integer;

    // parameters of last loaded file
    fFileSize : integer;
    fFileDate : integer;
    fFileAttr : integer;
    procedure ClearInfo;
   public
    constructor Create;
    destructor Destroy; override;

    procedure LoadFromFile (const AFileName : string);
    // Load MpegHeader and ID3Tag from file AFileName

    procedure LoadFromStream (AStream : TStream; AStreamSz : integer);
    // Load MpegHeader and ID3Tag in AStream
    // (max AStreamSz bytes from curren position)

    property Mpg : TMpegFrameHdr read fMpg;
    property ID3 : TID3v1Tag read fID3;

    property FileSize : integer read fFileSize;
    // Size of last loaded file

    property XingVBR : boolean read fXingVBR;
    // True if Xing VBR compression

    property MpegDuration : integer read fMpegDuration;
    // Sound duration (in ms) calculated from Mpg information

    property MaxGapBeforeMpeg : integer read fMaxGapBeforeMpeg write fMaxGapBeforeMpeg;
    // Max gap from file beginning to first MPEG frame.
    // If there are no Mpeg frame synch inside this gap,
    // then we assume that this is not MPEG file.
    // By default 10K. You may increase it, but too big
    // value will slow down scanning of big non-MPEG files.
  end;

  { Class TID3v2 }
  TID3v2 = class(TObject)
    private
      { Private declarations }
      FExists: Boolean;
      FVersionID: Byte;
      FSize: Integer;
      FTitle: string;
      FArtist: string;
      FAlbum: string;
      FTrack: Byte;
      FYear: string;
      FGenre: string;
      FComment: string;
      procedure FSetTitle(const NewTitle: string);
      procedure FSetArtist(const NewArtist: string);
      procedure FSetAlbum(const NewAlbum: string);
      procedure FSetTrack(const NewTrack: Byte);
      procedure FSetYear(const NewYear: string);
      procedure FSetGenre(const NewGenre: string);
      procedure FSetComment(const NewComment: string);
    public
      { Public declarations }
      constructor Create;                                     { Create object }
      procedure ResetData;                                   { Reset all data }
      function ReadFromFile(const FileName: string): Boolean;      { Load tag }
      function RemoveFromFile(const FileName: string): Boolean;  { Delete tag }
      function SaveToFile(const FileName: string): Boolean;        { Save tag }
      property Exists: Boolean read FExists;              { True if tag found }
      property VersionID: Byte read FVersionID;                { Version code }
      property Size: Integer read FSize;                     { Total tag size }
      property Title: string read FTitle write FSetTitle;        { Song title }
      property Artist: string read FArtist write FSetArtist;    { Artist name }
      property Album: string read FAlbum write FSetAlbum;       { Album title }
      property Track: Byte read FTrack write FSetTrack;        { Track number }
      property Year: string read FYear write FSetYear;         { Release year }
      property Genre: string read FGenre write FSetGenre;        { Genre name }
      property Comment: string read FComment write FSetComment;     { Comment }
  end;

  { Class TWMAfile }
  TWMAfile = class(TObject)
    private
      { Private declarations }
      FValid: Boolean;
      FFileSize: Integer;
      FChannelModeID: Byte;
      FSampleRate: Integer;
      FDuration: Double;
      FBitRate: Integer;
      FTitle: WideString;
      FArtist: WideString;
      FAlbum: WideString;
      FTrack: Integer;
      FYear: WideString;
      FGenre: WideString;
      FComment: WideString;
      procedure FResetData;
      function FGetChannelMode: string;
    public
      { Public declarations }
      constructor Create;                                     { Create object }
      function ReadFromFile(const FileName: string): Boolean;     { Load data }
      property Valid: Boolean read FValid;               { True if valid data }
      property FileSize: Integer read FFileSize;          { File size (bytes) }
      property ChannelModeID: Byte read FChannelModeID;   { Channel mode code }
      property ChannelMode: string read FGetChannelMode;  { Channel mode name }
      property SampleRate: Integer read FSampleRate;       { Sample rate (hz) }
      property Duration: Double read FDuration;          { Duration (seconds) }
      property BitRate: Integer read FBitRate;              { Bit rate (kbit) }
      property Title: WideString read FTitle;                    { Song title }
      property Artist: WideString read FArtist;                 { Artist name }
      property Album: WideString read FAlbum;                    { Album name }
      property Track: Integer read FTrack;                     { Track number }
      property Year: WideString read FYear;                            { Year }
      property Genre: WideString read FGenre;                    { Genre name }
      property Comment: WideString read FComment;                   { Comment }
  end;

var
 ID3Genres: TStringList;

const
 ID3v1TagLen = 128;

implementation

uses
 Windows, Messages;

const
  { ID3v2 tag ID }
  ID3V2_ID = 'ID3';

  { Max. number of supported tag frames }
  ID3V2_FRAME_COUNT = 7;

  { Names of supported tag frames }
  ID3V2_FRAME: array [1..ID3V2_FRAME_COUNT] of string =
    ('TIT2', 'TPE1', 'TALB', 'TRCK', 'TYER', 'TCON', 'COMM');

  { Object IDs }
  WMA_HEADER_ID =
    #48#38#178#117#142#102#207#17#166#217#0#170#0#98#206#108;
  WMA_FILE_PROPERTIES_ID =
    #161#220#171#140#71#169#207#17#142#228#0#192#12#32#83#101;
  WMA_STREAM_PROPERTIES_ID =
    #145#7#220#183#183#169#207#17#142#230#0#192#12#32#83#101;
  WMA_CONTENT_DESCRIPTION_ID =
    #51#38#178#117#142#102#207#17#166#217#0#170#0#98#206#108;
  WMA_EXTENDED_CONTENT_DESCRIPTION_ID =
    #64#164#208#210#7#227#210#17#151#240#0#160#201#94#168#80;

  { Max. number of supported comment fields }
  WMA_FIELD_COUNT = 7;

  { Names of supported comment fields }
  WMA_FIELD_NAME: array [1..WMA_FIELD_COUNT] of WideString =
    ('WM/TITLE', 'WM/AUTHOR', 'WM/ALBUMTITLE', 'WM/TRACK', 'WM/YEAR',
     'WM/GENRE', 'WM/DESCRIPTION');

  { Max. number of characters in tag field }
  WMA_MAX_STRING_SIZE = 250;

type
  { ID3v2 frame header }
  FrameHeader = record
    ID: array [1..4] of Char;                                      { Frame ID }
    Size: Integer;                                    { Size excluding header }
    Flags: Word;                                                      { Flags }
  end;

  { ID3v2 header data - for internal use }
  TagInfo = record
    { Real structure of ID3v2 header }
    ID: array [1..3] of Char;                                  { Always "ID3" }
    Version: Byte;                                           { Version number }
    Revision: Byte;                                         { Revision number }
    Flags: Byte;                                               { Flags of tag }
    Size: array [1..4] of Byte;                   { Tag size excluding header }
    { Extended data }
    FileSize: Integer;                                    { File size (bytes) }
    Frame: array [1..ID3V2_FRAME_COUNT] of string;  { Information from frames }
  end;

  { Object ID }
  ObjectID = array [1..16] of Char;

  { Tag data }
  TagData = array [1..WMA_FIELD_COUNT] of WideString;

  { File data - for internal use }
  FileData = record
    FileSize: Integer;                                    { File size (bytes) }
    MaxBitRate: Integer;                                { Max. bit rate (bps) }
    Channels: Word;                                      { Number of channels }
    SampleRate: Integer;                                   { Sample rate (hz) }
    ByteRate: Integer;                                            { Byte rate }
    Tag: TagData;                                       { WMA tag information }
  end;

constructor TMpegFrameHdr.Create;
 begin
  inherited;
  fOk := False;
  FillChar (fHdrBuf, SizeOf (fHdrBuf), 0);
 end; { of constructor TMpegFrameHdr.Create
--------------------------------------------------------------}

function TMpegFrameHdr.Load (const ABuf; ABufSz : integer) : boolean;
 begin
  if ABufSz >= MPegFrameHdrSz then begin
     Move (ABuf, fHdrBuf, MPegFrameHdrSz);
     // optimization - precalculate some fields
     fVersion := GetVersion;
     fKBitRate := GetKBitRate; // depends on Version & Layer !!
     fSamplingRate := GetSamplingRate; // depends on Version !!
     fFrameLen := GetFrameLen; // depends on Layer, KBitRate, SamplingRate & Padding !!
     fOk := GetOk; // Depends on nearly all fields
     Result := Ok;
    end
   else Result := False;
 end; { of function TMpegFrameHdr.Load
--------------------------------------------------------------}

function TMpegFrameHdr.Locate (const ABuf; ABufSz : integer) : integer;
 var
  p : PChar;
 begin
  Result := 0;
  p := @ABuf;
  while Result <= (ABufSz - MPegFrameHdrSz) do
   if (p^ = #$FF) and ((ord ((p + 1)^) and $E0) = $E0)
      and Load (p^, ABufSz - Result)
    then EXIT
    else begin
      inc (p);
      inc (Result);
     end;
  Result := -1;
 end; { of function TMpegFrameHdr.Locate
--------------------------------------------------------------}

function TMpegFrameHdr.GetOk : boolean;
 begin
  Result := Sync and (Version <> MPEG_VER_ERROR)
    and (Layer <> MPEG_LAYER_ERROR)
    and (KBitRate <> MPEG_BITRATE_ERROR)
    and (SamplingRate <> MPEG_SAMPLINGRATE_ERROR);
 end; { of function TMpegFrameHdr.GetOk
--------------------------------------------------------------}

function TMpegFrameHdr.GetSync : boolean;
 begin
  Result := (fHdrBuf [0] = $FF) and ((fHdrBuf [1] and $E0) = $E0);
 end; { of function TMpegFrameHdr.GetSync
--------------------------------------------------------------}

function TMpegFrameHdr.GetVersion : integer;
 begin
  case (fHdrBuf [1] ShR 3) and 3 of
    0: Result := MPEG_VER_2_5;
    2: Result := MPEG_VER_2;
    3: Result := MPEG_VER_1;
    else Result := MPEG_VER_ERROR;
   end;
 end; { of function TMpegFrameHdr.GetVersion
--------------------------------------------------------------}

function TMpegFrameHdr.GetVersionStr : string;
 begin
  Result := MpegVerStr [Version];
 end; { of function TMpegFrameHdr.GetVersionStr
--------------------------------------------------------------}

function TMpegFrameHdr.GetLayer : integer;
 begin
  Result := (4 - (fHdrBuf [1] ShR 1) and 3) and 3;
 end; { of function TMpegFrameHdr.GetLayer
--------------------------------------------------------------}

function TMpegFrameHdr.GetLayerStr : string;
 begin
  Result := MpegLayerStr [Layer];
 end; { of function TMpegFrameHdr.GetLayerStr
--------------------------------------------------------------}

function TMpegFrameHdr.GetCRCExists : boolean;
 begin
  Result := (fHdrBuf [1] and 1) = 0;
 end; { of function TMpegFrameHdr.GetCRCExists
--------------------------------------------------------------}

function TMpegFrameHdr.GetKBitRate : integer;
 const
  MpegKBitRate : array [1 .. 3, 1 .. 3, 0 .. 15] of integer = (
  (//MPEG 1
   (0, 32, 64, 96,128,160,192,224,256,288,320,352,384,416,448,MPEG_BITRATE_ERROR), //Layer I
   (0, 32, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,384,MPEG_BITRATE_ERROR), //Layer II
   (0, 32, 40, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,MPEG_BITRATE_ERROR)  //Layer III
  ),
  (//MPEG 2
   (0, 32, 48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,MPEG_BITRATE_ERROR), //Layer I)
   (0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,MPEG_BITRATE_ERROR), //Layer II
   (0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,MPEG_BITRATE_ERROR)  //Layer III
  ),
  (//MPEG 2.5
   (0, 32, 48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,MPEG_BITRATE_ERROR), //Layer I)
   (0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,MPEG_BITRATE_ERROR), //Layer II
   (0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,MPEG_BITRATE_ERROR)  //Layer III
  ));
 begin
  if (Version <> MPEG_VER_ERROR) and (Layer <> MPEG_LAYER_ERROR)
   then Result := MpegKBitRate [Version, Layer, (fHdrBuf [2] ShR 4) and $F]
   else Result := MPEG_BITRATE_ERROR
 end; { of function TMpegFrameHdr.GetKBitRate
--------------------------------------------------------------}

function TMpegFrameHdr.GetBitRate : integer;
 begin
  Result := KBitRate * 1000;
 end; { of function TMpegFrameHdr.GetBitRate
--------------------------------------------------------------}

function TMpegFrameHdr.GetSamplingRate : integer;
 const
  MpegSamplingRate : array [1 .. 3, 0 .. 3] of integer = (
    (44100, 48000, 32000, MPEG_SAMPLINGRATE_ERROR), //MPEG 1
    (22050, 24000, 16000, MPEG_SAMPLINGRATE_ERROR), //MPEG 2
    (32000, 16000,  8000, MPEG_SAMPLINGRATE_ERROR)  //MPEG 2.5
   );
 begin
  if Version <> MPEG_VER_ERROR
   then Result := MpegSamplingRate [Version, (fHdrBuf [2] ShR 2) and 3]
   else Result := MPEG_SAMPLINGRATE_ERROR;
 end; { of function TMpegFrameHdr.GetSamplingRate
--------------------------------------------------------------}

function TMpegFrameHdr.GetPadding : boolean;
 begin
  Result := ((fHdrBuf [2] ShR 1) and 1) = 1;
 end; { of function TMpegFrameHdr.GetPadding
--------------------------------------------------------------}

function TMpegFrameHdr.GetPaddingLen : integer;
 begin
  if Padding
   then
     if Layer = MPEG_LAYER_I
      then Result := 4
      else Result := 1
   else Result := 0;
 end; { of function TMpegFrameHdr.GetPaddingLen
--------------------------------------------------------------}

function TMpegFrameHdr.GetChannelMode : integer;
 begin
  Result := (fHdrBuf [3] ShR 6) and 3;
 end; { of function TMpegFrameHdr.GetChannelMode
--------------------------------------------------------------}

function TMpegFrameHdr.GetChannelModeStr : string;
 begin
  Result := MpegChannelStr [ChannelMode];
 end; { of function TMpegFrameHdr.GetChannelModeStr
--------------------------------------------------------------}

function TMpegFrameHdr.GetChannelModeExt : integer;
 begin
  Result := (fHdrBuf [3] ShR 4) and 3;
 end; { of function TMpegFrameHdr.GetChannelModeExt
--------------------------------------------------------------}

function TMpegFrameHdr.GetChannelModeExtStr : string;
 const
  MpegChannelExtStr : array [0 .. 1, 0 .. 3] of string = (
   ('Bands 4-31', 'Bands 8-31', 'Bands 12-31', 'Bands 16-31'),
   ('', 'Intensity', 'MS', 'MS/Intensity'));
 begin
  if ChannelMode = MPEG_CHANNEL_JOINTSTEREO
   then
     if Layer = MPEG_LAYER_III
      then Result := MpegChannelExtStr [1, ChannelModeExt]
      else Result := MpegChannelExtStr [0, ChannelModeExt]
   else Result := '';
 end; { of function TMpegFrameHdr.GetChannelModeExtStr
--------------------------------------------------------------}

function TMpegFrameHdr.GetCopyright : boolean;
 begin
  Result := ((fHdrBuf [3] ShR 3) and 1) = 1;
 end; { of function TMpegFrameHdr.GetCopyright
--------------------------------------------------------------}

function TMpegFrameHdr.GetOriginal : boolean;
 begin
  Result := ((fHdrBuf [3] ShR 2) and 1) = 1;
 end; { of function TMpegFrameHdr.GetOriginal
--------------------------------------------------------------}

function TMpegFrameHdr.GetEmphasis : integer;
 begin
  Result := fHdrBuf [3] and 3;
 end; { of function TMpegFrameHdr.GetEmphasis
--------------------------------------------------------------}

function TMpegFrameHdr.GetEmphasisStr : string;
 begin
  Result := MpegEmphasisStr [Emphasis];
 end; { of function TMpegFrameHdr.GetEmphasisStr
--------------------------------------------------------------}

function TMpegFrameHdr.GetFrameLen : integer;
{  384 samples/frame for Layer I
  1152 samples/frame for Layer II and Layer III
  8 bits Mpeg 1, 16 bits Mpeg 2 & 2.5}
 begin
  if SamplingRate > 0
   then
     if Version = MPEG_VER_1
      then
        if Layer = MPEG_LAYER_I
         then Result := Trunc ((48 * BitRate) / SamplingRate) + PaddingLen
         else Result := Trunc ((144 * BitRate) / SamplingRate) + PaddingLen
      else
        if Layer = MPEG_LAYER_I
         then Result := Trunc ((24 * BitRate) / SamplingRate) + PaddingLen
         else Result := Trunc ((72 * BitRate) / SamplingRate) + PaddingLen
   else Result := 0;
 end; { of function TMpegFrameHdr.GetFrameLen
--------------------------------------------------------------}

constructor TID3v1Tag.Create;
 begin
  inherited;
  fOk := false;
  fModified := false;
 end; { of constructor TID3v1Tag.Create
--------------------------------------------------------------}

procedure TID3v1Tag.Setv1 (b : boolean);
 begin
  if fv1 <> b then begin
    fv1 := b;
    fModified := True;
   end;
 end; { procedure TID3v1Tag.Setv1
--------------------------------------------------------------}

procedure TID3v1Tag.SetField (AIdx : integer; const s : string);
 procedure SetIt (const ANewValue : string; AMaxLen : integer; var AField : string);
  begin
   if copy (ANewValue, 1, AMaxLen) <> AField then begin
     AField := copy (ANewValue, 1, AMaxLen);
     fModified := True;
    end;
  end;
 begin
  case AIdx of
    1: SetIt (s, 30, fTitle);
    2: SetIt (s, 30, fArtist);
    3: SetIt (s, 4, fYear);
    4: SetIt (s, 30, fAlbum);
    5: if fv1
        then SetIt (s, 30, fComment)
        else SetIt (s, 28, fComment);
   end;
 end; { of procedure TID3v1Tag.SetTitle
--------------------------------------------------------------}

procedure TID3v1Tag.SetGenreID (AGenreID : integer);
 begin
  if fGenreID <> AGenreID then begin
    fGenreID := AGenreID;
    fModified := True;
   end;
 end; { of procedure TID3v1Tag.SetGenreID
--------------------------------------------------------------}

procedure TID3v1Tag.SetTrack (ATrack : integer);
 begin
  if fTrack <> ATrack then begin
    fTrack := ATrack;
    fv1 := fTrack = 0;
    fModified := True;
   end;
 end; { of procedure TID3v1Tag.SetTrack
--------------------------------------------------------------}

procedure TID3v1Tag.Load (const ABuf; ABufSz : integer);
 var
  b : array [0 .. ID3v1TagLen - 1] of byte absolute ABuf;
  GenreIdx : integer;
 function BufStr (APos, ALen : integer) : string;
  begin
   SetString (Result, nil, ALen);
   Move (b [APos], Result [1], ALen);
   if Pos (#0, Result) > 0
    then Result := copy (Result, 1, Pos (#0, Result) - 1);
   Result := TrimRight (Result);
  end;
 begin
  fOk := false;
  fModified := false;
  if ABufSz < ID3v1TagLen
   then EXIT;
    if BufStr (0, 3) = 'TAG' then begin
      fOk := true;
      fTitle := BufStr (3, 30);
      fArtist:= BufStr (33, 30);
      fAlbum:= BufStr (63, 30);
      fYear:= BufStr (93, 4);
      fComment:= BufStr (97, 30);
      if (b [97 + 28] = 0) and (b [97 + 29] <> 0) then begin
         fv1 := false;
         fTrack := b [97 + 29];
        end
       else begin
         fv1 := true;
         fTrack := 0;
        end;
      fGenreID := b [127];
      GenreIdx := ID3Genres.IndexOfObject (TObject (fGenreID));
      if GenreIdx >= 0
       then fGenre := ID3Genres [GenreIdx]
       else fGenre := '';
     end;
 end; { of procedure TID3v1Tag.Load
--------------------------------------------------------------}

procedure TID3v1Tag.LoadFromStream (AStream : TStream; AStreamSz : integer);
 var
  Buf : array [0 .. ID3v1TagLen - 1] of byte;
 begin
  if AStreamSz >= ID3v1TagLen then begin
     AStream.Position := AStream.Position + AStreamSz - ID3v1TagLen;
     AStream.ReadBuffer (Buf, ID3v1TagLen);
     Load (Buf, ID3v1TagLen);
    end
   else fOk := false;
 end; { of procedure TID3v1Tag.LoadFromStream
--------------------------------------------------------------}

procedure TID3v1Tag.LoadFromFile (const AFileName : string);
 var
  f : TStream;
 begin
  f := TFileStream.Create (AFileName, fmOpenRead or fmShareDenyNone);
  try
     LoadFromStream (f, f.Size);
    finally f.Free;
   end;
 end; { of procedure TID3v1Tag.LoadFromFile
--------------------------------------------------------------}

procedure TID3v1Tag.Save (var ABuf; ABufSz : integer);
 var
  b : array [0 .. ID3v1TagLen - 1] of byte absolute ABuf;
 procedure StrBuf (APos, ALen : integer; const AStr : string);
  var
   Len : integer;
  begin
   Len := Length (AStr);
   if Len > ALen
    then Len := ALen;
   Move (AStr [1], b [APos], Len);
   if Len < ALen
    then FillChar (b [APos + Len], ALen - Len, $20);
  end;
 begin
  if ABufSz < ID3v1TagLen
   then raise Exception.Create ('TID3v1Tag.Save: Buffer too small');
  // fOk ?!!
  StrBuf (0, 3, 'TAG');
  StrBuf (3, 30, fTitle);
  StrBuf (33, 30, fArtist);
  StrBuf (63, 30, fAlbum);
  StrBuf (93, 4, fYear);
  StrBuf (97, 30, fComment);
  if not fv1 then begin
    b [97 + 28] := 0;
    b [97 + 29] := fTrack;
   end;
  b [127] := fGenreID;
  fModified := false;
 end; { of procedure TID3v1Tag.Save
--------------------------------------------------------------}

procedure TID3v1Tag.SaveToStream (AStream : TStream);
 var
  Buf : array [0 .. ID3v1TagLen - 1] of byte;
 begin
  Save (Buf, ID3v1TagLen);
  AStream.WriteBuffer (Buf, ID3v1TagLen);
 end; { of procedure TID3v1Tag.SaveToStream
--------------------------------------------------------------}

procedure TID3v1Tag.SaveToFile (const AFileName : string);
 var
  f : TStream;
  s : string;
 begin
  f := TFileStream.Create (AFileName, fmOpenReadWrite or fmShareDenyNone);
  try
     if f.Size >= ID3v1TagLen then begin
        f.Position := f.Size - ID3v1TagLen;
        SetString (s, nil, 3);
        f.ReadBuffer (s [1], 3);
       end
      else s := 'NON';
     if s = 'TAG'
      then f.Position := f.Size - ID3v1TagLen
      else f.Position := f.Size;
     SaveToStream (f);
    finally f.Free;
   end;
 end; { of procedure TID3v1Tag.SaveToFile
--------------------------------------------------------------}

procedure TID3v1Tag.DeleteFromFile (const AFileName : string);
 var
  f : TStream;
  s : string;
 begin
  f := TFileStream.Create (AFileName, fmOpenReadWrite or fmShareExclusive);
  try
     if f.Size >= ID3v1TagLen then begin
        f.Position := f.Size - ID3v1TagLen;
        SetString (s, nil, 3);
        f.ReadBuffer (s [1], 3);
        if s = 'TAG' then begin
          f.Size := f.Size - ID3v1TagLen;
         end;
       end;
    finally f.Free;
   end;
 end; { of procedure TID3v1Tag.DeleteFromFile
--------------------------------------------------------------}


{=============================================================}
{======================= TAudioInfo ==========================}

constructor TAudioInfo.Create;
 begin
  inherited;
  fMpg := TMpegFrameHdr.Create;
  fID3 := TID3v1Tag.Create;
  fMaxGapBeforeMpeg := 10 * 1024;
 end; { of constructor TAudioInfo.Create
--------------------------------------------------------------}

destructor TAudioInfo.Destroy;
 begin
  fID3.Free;
  fMpg.Free;
  inherited;
 end; { of destructor TAudioInfo.Destroy
--------------------------------------------------------------}

procedure TAudioInfo.ClearInfo;
 begin
  fMpg.Ok := false;
  fFirstMpegFramePos := -1;
  fXingVBR := false;
  fID3.Ok := false;
  fFileSize := -1;
  fFileDate := -1;
  fFileAttr := 0;
 end; { of procedure TAudioInfo.ClearInfo
--------------------------------------------------------------}

procedure TAudioInfo.LoadFromFile (const AFileName : string);
 var
  f : TStream;
 begin
  ClearInfo;
  if length (Trim (AFileName)) > 0 then begin
    f := TFileStream.Create (AFileName, fmOpenRead or fmShareDenyNone);
    try
       LoadFromStream (f, f.Size);
       fFileSize := f.Size;
      finally f.Free;
     end;
    fFileDate := FileAge (AFileName);
    fFileAttr := FileGetAttr (AFileName);
   end;
 end; { of procedure TAudioInfo.LoadFromFile
--------------------------------------------------------------}

procedure TAudioInfo.LoadFromStream (AStream : TStream; AStreamSz : integer);
 const
  BufSz = 128; // must be > MpegFrameHdrSz !!!
 var
  Buf : array [0 .. BufSz - 1] of byte;
  Sz, Pos, StartPos, Off : integer;
  XingOff, XingFrames : integer;
  MpegLen : integer;
  MpegHdr2 : TMpegFrameHdr;
 begin
  // Clear it all. If we'll fail, then all "Ok" flags will be False.
  ClearInfo;

  // Save starting position in stream (may be it <> 0)
  StartPos := AStream.Position;

  // protection against dummy
  if AStreamSz > AStream.Size - StartPos
   then AStreamSz := AStream.Size - StartPos;

  // find first MPEG frame
  Pos := 0; // position relative to StartPos
  Off := 0; // for assembling file slices
  while (Pos < fMaxGapBeforeMpeg) and (Pos < AStreamSz) do begin
    Sz := BufSz - Off;
    if Pos + Sz >= AStreamSz
     then Sz := AStreamSz - Pos;
    AStream.ReadBuffer (Buf [Off], Sz);
    fFirstMpegFramePos := fMpg.Locate (Buf, Sz + Off);
    if fFirstMpegFramePos >= 0 then begin // Urrra! Bingo !
      // convert "pos in slice" to "offset from StartPos"
      inc (fFirstMpegFramePos, Pos - Off);
      BREAK;
     end;
    inc (Pos, Sz); // update current read pos (rel to StartPos)
    Off := MpegFrameHdrSz; // after first file slice,
    // all next slices will be shifted in Buf
    // for continues processing emulation
    Move (Buf [BufSz - MpegFrameHdrSz], Buf, MpegFrameHdrSz);
    // - shift part of previous slice to Buf's beginnig
   end;

  if fMpg.Ok then begin // Valid MPEG frame found
    // Check next for be sure...
    Pos := fFirstMpegFramePos + fMpg.FrameLen;
    if Pos + MpegFrameHdrSz < AStreamSz then begin
      AStream.Position := StartPos + Pos;
      AStream.ReadBuffer (Buf, MpegFrameHdrSz);
      MpegHdr2 := TMpegFrameHdr.Create;
      try
         MpegHdr2.Load (Buf, MpegFrameHdrSz);
         fMpg.Ok := MpegHdr2.Ok;
        finally MpegHdr2.Free;
       end;
     end;
   end;

  if fMpg.Ok then begin
     // Is this Xing VBR file ?
     if (fMpg.Version = MPEG_VER_1)
      then
        if fMpg.ChannelMode <> MPEG_CHANNEL_SINGLE
         then XingOff := 32 + 4
         else XingOff := 17 + 4
      else
        if fMpg.ChannelMode <> MPEG_CHANNEL_SINGLE
         then XingOff := 17 + 4
         else XingOff := 9 + 4;
     if fFirstMpegFramePos + XingOff + 12 < AStreamSz then begin
       AStream.Position := StartPos + fFirstMpegFramePos + XingOff;
       AStream.ReadBuffer (Buf, 12);
       if (Buf [0] = Ord ('X')) and (Buf [1] = Ord ('i')) and
          (Buf [2] = Ord ('n')) and (Buf [3] = Ord ('g')) then begin
         fXingVBR := true;
         if (Buf [7] and 1) <> 0 // frames number specified
          then XingFrames := (((((Buf [8] ShL 8) + Buf [9]) ShL 8)
                 + Buf [10]) ShL 8) + Buf [11]
          else XingFrames := 0;
         if fMpg.SamplingRate > 0
          then fMpegDuration := Round ((1152 / fMpg.SamplingRate) * XingFrames) * 1000
          else fMpegDuration := 0;
        end;
      end;
     Pos := fFirstMpegFramePos + fMpg.FrameLen;
    end
   else begin // not fMpeg.Ok
     // May be this is unknown MPEG-format.
     // Reset to StartPos for ID3-tags finding.
     Pos := 0;
    end;
  AStream.Position := StartPos + Pos;


  // try to load ID3v1.1 tag
  fID3.LoadFromStream (AStream, AStreamSz - Pos);

  // calculate Duration based on MpegHdr
  if not fXingVBR then begin
    MpegLen := AStreamSz - fFirstMpegFramePos;
    if fID3.Ok
     then dec (MpegLen, ID3v1TagLen);
    if fMpg.BitRate <> MPEG_BITRATE_ERROR
     then fMpegDuration := (MpegLen * 8 div fMpg.BitRate) * 1000 { ms}
     else fMpegDuration := 0;
   end;
 end; { of procedure TAudioInfo.LoadFromStream
--------------------------------------------------------------}

{ ********************* Auxiliary functions & procedures ******************** }

function ReadHeader(const FileName: string; var Tag: TagInfo): Boolean;
var
  SourceFile: file;
  Transferred: Integer;
begin
  try
    Result := true;
    { Set read-access and open file }
    AssignFile(SourceFile, FileName);
    FileMode := 0;
    Reset(SourceFile, 1);
    { Read header and get file size }
    BlockRead(SourceFile, Tag, 10, Transferred);
    Tag.FileSize := FileSize(SourceFile);
    CloseFile(SourceFile);
    { if transfer is not complete }
    if Transferred < 10 then Result := false;
  except
    { Error }
    Result := false;
  end;
end;

{ --------------------------------------------------------------------------- }

function GetVersionID(const Tag: TagInfo): Byte;
begin
  { Get tag version from header }
  Result := Tag.Version;
end;

{ --------------------------------------------------------------------------- }

function GetTagSize(const Tag: TagInfo): Integer;
begin
  { Get total tag size }
  Result :=
    Tag.Size[1] * $200000 +
    Tag.Size[2] * $4000 +
    Tag.Size[3] * $80 +
    Tag.Size[4] + 10;
  if Result > Tag.FileSize then Result := 0;
end;

{ --------------------------------------------------------------------------- }

procedure SetTagItem(const ID, Data: string; var Tag: TagInfo);
var
  Iterator: Byte;
begin
  { Set tag item if supported frame found }
  for Iterator := 1 to ID3V2_FRAME_COUNT do
    if ID3V2_FRAME[Iterator] = ID then Tag.Frame[Iterator] := Data;
end;

{ --------------------------------------------------------------------------- }

function Swap32(const Figure: Integer): Integer;
var
  ByteArray: array [1..4] of Byte absolute Figure;
begin
  { Swap 4 bytes }
  Result :=
    ByteArray[1] * $1000000 +
    ByteArray[2] * $10000 +
    ByteArray[3] * $100 +
    ByteArray[4];
end;

{ --------------------------------------------------------------------------- }

procedure ReadFrames(const FileName: string; var Tag: TagInfo);
var
  SourceFile: file;
  Frame: FrameHeader;
  Data: array [1..250] of Char;
  DataPosition: Integer;
begin
  try
    { Set read-access, open file }
    AssignFile(SourceFile, FileName);
    FileMode := 0;
    Reset(SourceFile, 1);
    Seek(SourceFile, 10);
    while (FilePos(SourceFile) < GetTagSize(Tag)) and (not EOF(SourceFile)) do
    begin
      FillChar(Data, SizeOf(Data), 0);
      { Read frame header and check frame ID }
      BlockRead(SourceFile, Frame, 10);
      if not (Frame.ID[1] in ['A'..'Z']) then break;
      DataPosition := FilePos(SourceFile);
      { Read frame data and set tag item if frame supported }
      BlockRead(SourceFile, Data, Swap32(Frame.Size) mod SizeOf(Data));
      SetTagItem(Frame.ID, Data, Tag);
      Seek(SourceFile, DataPosition + Swap32(Frame.Size));
    end;
    CloseFile(SourceFile);
  except
  end;
end;

{ --------------------------------------------------------------------------- }

function ExtractTrack(const TrackString: string): Byte;
var
  Index, Value, Code: Integer;
begin
  { Extract track from string }
  Index := Pos('/', Trim(TrackString));
  if Index = 0 then Val(Trim(TrackString), Value, Code)
  else Val(Copy(Trim(TrackString), 1, Index - 1), Value, Code);
  if Code = 0 then Result := Value
  else Result := 0;
end;

{ --------------------------------------------------------------------------- }

function ExtractGenre(const GenreString: string): string;
begin
  { Extract genre from string }
  Result := Trim(GenreString);
  if Pos(')', Result) > 0 then Delete(Result, 1, LastDelimiter(')', Result));
end;

{ --------------------------------------------------------------------------- }

function RebuildFile(const FileName: string; NewTagData: TStream): Boolean;
var
  Tag: TagInfo;
  Source, Destination: TFileStream;
  BufferName: string;
begin
  { Rebuild file with old file data and new tag (optional) }
  Result := false;
  if (not FileExists(FileName)) or (FileSetAttr(FileName, 0) <> 0) then exit;
  if not ReadHeader(FileName, Tag) then exit;
  if (NewTagData = nil) and (Tag.ID <> ID3V2_ID) then exit;
  try
    { Create file streams }
    BufferName := FileName + '~';
    Source := TFileStream.Create(FileName, fmOpenRead	or fmShareExclusive);
    Destination := TFileStream.Create(BufferName, fmCreate);
    { Copy data blocks }
    if Tag.ID = ID3V2_ID then Source.Seek(GetTagSize(Tag), soFromBeginning);
    if NewTagData <> nil then Destination.CopyFrom(NewTagData, 0);
    Destination.CopyFrom(Source, Source.Size - Source.Position);
    { Free resources }
    Source.Free;
    Destination.Free;
    { Replace old file and delete temporary file }
    if (SysUtils.DeleteFile(FileName)) and (RenameFile(BufferName, FileName)) then
      Result := true
    else
      raise Exception.Create('');
  except
    { Access error }
    if FileExists(BufferName) then SysUtils.DeleteFile(BufferName);
  end;
end;

{ --------------------------------------------------------------------------- }

procedure BuildHeader(var Tag: TagInfo);
var
  Iterator, TagSize: Integer;
begin
  { Build tag header }
  Tag.ID := ID3V2_ID;
  Tag.Version := TAG_VERSION_2_3;
  Tag.Revision := 0;
  Tag.Flags := 0;
  TagSize := 0;
  for Iterator := 1 to ID3V2_FRAME_COUNT do
    if Tag.Frame[Iterator] <> '' then
      Inc(TagSize, Length(Tag.Frame[Iterator]) + 11);
  { Convert tag size }
  Tag.Size[1] := TagSize div $200000;
  Tag.Size[2] := TagSize div $4000;
  Tag.Size[3] := TagSize div $80;
  Tag.Size[4] := TagSize mod $80;
end;

{ ********************** Private functions & procedures ********************* }

procedure TID3v2.FSetTitle(const NewTitle: string);
begin
  { Set song title }
  FTitle := TrimRight(NewTitle);
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetArtist(const NewArtist: string);
begin
  { Set artist name }
  FArtist := TrimRight(NewArtist);
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetAlbum(const NewAlbum: string);
begin
  { Set album title }
  FAlbum := TrimRight(NewAlbum);
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetTrack(const NewTrack: Byte);
begin
  { Set track number }
  FTrack := NewTrack;
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetYear(const NewYear: string);
begin
  { Set release year }
  FYear := TrimRight(NewYear);
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetGenre(const NewGenre: string);
begin
  { Set genre name }
  FGenre := TrimRight(NewGenre);
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.FSetComment(const NewComment: string);
begin
  { Set comment }
  FComment := TrimRight(NewComment);
end;

{ ********************** Public functions & procedures ********************** }

constructor TID3v2.Create;
begin
  { Create object }
  inherited;
  ResetData;
end;

{ --------------------------------------------------------------------------- }

procedure TID3v2.ResetData;
begin
  { Reset all variables }
  FExists := false;
  FVersionID := 0;
  FSize := 0;
  FTitle := '';
  FArtist := '';
  FAlbum := '';
  FTrack := 0;
  FYear := '';
  FGenre := '';
  FComment := '';
end;

{ --------------------------------------------------------------------------- }

function TID3v2.ReadFromFile(const FileName: string): Boolean;
var
  Tag: TagInfo;
begin
  { Reset data and load header from file to variable }
  ResetData;
  Result := ReadHeader(FileName, Tag);
  { Process data if loaded and header valid }
  if (Result) and (Tag.ID = ID3V2_ID) then
  begin
    FExists := true;
    { Fill properties with header data }
    FVersionID := GetVersionID(Tag);
    FSize := GetTagSize(Tag);
    { Get information from frames if version supported }
    if (FVersionID = TAG_VERSION_2_3) and (FSize > 0) then
    begin
      ReadFrames(FileName, Tag);
      { Fill properties with data from frames }
      FTitle := Trim(Tag.Frame[1]);
      FArtist := Trim(Tag.Frame[2]);
      FAlbum := Trim(Tag.Frame[3]);
      FTrack := ExtractTrack(Tag.Frame[4]);
      FYear := Trim(Tag.Frame[5]);
      FGenre := ExtractGenre(Tag.Frame[6]);
      FComment := Trim(Copy(Tag.Frame[7], 5, Length(Tag.Frame[7]) - 4));
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

function TID3v2.RemoveFromFile(const FileName: string): Boolean;
begin
  { Remove tag from file }
  Result := RebuildFile(FileName, nil);
end;

{ --------------------------------------------------------------------------- }

function TID3v2.SaveToFile(const FileName: string): Boolean;
var
  NewTagData: TStringStream;
  Tag: TagInfo;
  Iterator, FrameSize: Integer;
begin
  { Build tag and save to file }
  NewTagData := TStringStream.Create('');
  { Prepare tag data }
  FillChar(Tag, SizeOf(Tag), 0);
  Tag.Frame[1] := FTitle;
  Tag.Frame[2] := FArtist;
  Tag.Frame[3] := FAlbum;
  if FTrack > 0 then Tag.Frame[4] := IntToStr(FTrack);
  Tag.Frame[5] := FYear;
  Tag.Frame[6] := FGenre;
  if FComment <> '' then Tag.Frame[7] := 'eng' + #0 + FComment;
  { Build and write tag header }
  BuildHeader(Tag);
  NewTagData.Write(Tag, 10);
  { Build and write tag frames }
  for Iterator := 1 to ID3V2_FRAME_COUNT do
    if Tag.Frame[Iterator] <> '' then
    begin
      NewTagData.WriteString(ID3V2_FRAME[Iterator]);
      FrameSize := Swap32(Length(Tag.Frame[Iterator]) + 1);
      NewTagData.Write(FrameSize, SizeOf(FrameSize));
      NewTagData.WriteString(#0#0#0 + Tag.Frame[Iterator]);
    end;
  { Rebuild file with new tag }
  Result := RebuildFile(FileName, NewTagData);
  NewTagData.Free;
end;

{ ********************* Auxiliary functions & procedures ******************** }

function ReadFieldString(const Source: TStream; DataSize: Word): WideString;
var
  Iterator, StringSize: Integer;
  FieldData: array [1..WMA_MAX_STRING_SIZE * 2] of Byte;
begin
  { Read field data and convert to unicode string }
  Result := '';
  StringSize := DataSize div 2;
  if StringSize > WMA_MAX_STRING_SIZE then StringSize := WMA_MAX_STRING_SIZE;
  Source.ReadBuffer(FieldData, StringSize * 2);
  Source.Seek(DataSize - StringSize * 2, soFromCurrent);
  for Iterator := 1 to StringSize do
    Result := Result +
      WideChar(FieldData[Iterator * 2 - 1] + (FieldData[Iterator * 2] shl 8));
end;

{ --------------------------------------------------------------------------- }

procedure ReadTagStandard(const Source: TStream; var Tag: TagData);
var
  Iterator: Integer;
  FieldSize: array [1..5] of Word;
  FieldValue: WideString;
begin
  { Read standard tag data }
  Source.ReadBuffer(FieldSize, SizeOf(FieldSize));
  for Iterator := 1 to 5 do
    if FieldSize[Iterator] > 0 then
    begin
      { Read field value }
      FieldValue := ReadFieldString(Source, FieldSize[Iterator]);
      { Set corresponding tag field if supported }
      case Iterator of
        1: Tag[1] := FieldValue;
        2: Tag[2] := FieldValue;
        4: Tag[7] := FieldValue;
      end;
    end;
end;

{ --------------------------------------------------------------------------- }

procedure ReadTagExtended(const Source: TStream; var Tag: TagData);
var
  Iterator1, Iterator2, FieldCount, DataSize, DataType: Word;
  FieldName, FieldValue: WideString;
begin
  { Read extended tag data }
  Source.ReadBuffer(FieldCount, SizeOf(FieldCount));
  for Iterator1 := 1 to FieldCount do
  begin
    { Read field name }
    Source.ReadBuffer(DataSize, SizeOf(DataSize));
    FieldName := ReadFieldString(Source, DataSize);
    { Read value data type }
    Source.ReadBuffer(DataType, SizeOf(DataType));
    { Read field value only if string }
    if DataType = 0 then
    begin
      Source.ReadBuffer(DataSize, SizeOf(DataSize));
      FieldValue := ReadFieldString(Source, DataSize);
    end
    else
      Source.Seek(DataSize, soFromCurrent);
    { Set corresponding tag field if supported }
    for Iterator2 := 1 to WMA_FIELD_COUNT do
      if UpperCase(Trim(FieldName)) = WMA_FIELD_NAME[Iterator2] then
        Tag[Iterator2] := FieldValue;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure ReadObject(const ID: ObjectID; Source: TStream; var Data: FileData);
begin
  { Read data from header object if supported }
  if ID = WMA_FILE_PROPERTIES_ID then
  begin
    { Read file properties }
    Source.Seek(80, soFromCurrent);
    Source.ReadBuffer(Data.MaxBitRate, SizeOf(Data.MaxBitRate));
  end;
  if ID = WMA_STREAM_PROPERTIES_ID then
  begin
    { Read stream properties }
    Source.Seek(60, soFromCurrent);
    Source.ReadBuffer(Data.Channels, SizeOf(Data.Channels));
    Source.ReadBuffer(Data.SampleRate, SizeOf(Data.SampleRate));
    Source.ReadBuffer(Data.ByteRate, SizeOf(Data.ByteRate));
  end;
  if ID = WMA_CONTENT_DESCRIPTION_ID then
  begin
    { Read standard tag data }
    Source.Seek(4, soFromCurrent);
    ReadTagStandard(Source, Data.Tag);
  end;
  if ID = WMA_EXTENDED_CONTENT_DESCRIPTION_ID then
  begin
    { Read extended tag data }
    Source.Seek(4, soFromCurrent);
    ReadTagExtended(Source, Data.Tag);
  end;
end;

{ --------------------------------------------------------------------------- }

function ReadData(const FileName: string; var Data: FileData): Boolean;
var
  Source: TFileStream;
  ID: ObjectID;
  Iterator, ObjectCount, ObjectSize, Position: Integer;
begin
  { Read file data }
  try
    Source := TFileStream.Create(FileName, fmOpenRead	or fmShareDenyWrite);
    Data.FileSize := Source.Size;
    { Check for existing header }
    Source.ReadBuffer(ID, SizeOf(ID));
    if ID = WMA_HEADER_ID then
    begin
      Source.Seek(8, soFromCurrent);
      Source.ReadBuffer(ObjectCount, SizeOf(ObjectCount));
      Source.Seek(2, soFromCurrent);
      { Read all objects in header and get needed data }
      for Iterator := 1 to ObjectCount do
      begin
        Position := Source.Position;
        Source.ReadBuffer(ID, SizeOf(ID));
        Source.ReadBuffer(ObjectSize, SizeOf(ObjectSize));
        ReadObject(ID, Source, Data);
        Source.Seek(Position + ObjectSize, soFromBeginning);
      end;
    end;
    Source.Free;
    Result := true;
  except
    Result := false;
  end;
end;

{ --------------------------------------------------------------------------- }

function IsValid(const Data: FileData): Boolean;
begin
  { Check for data validity }
  Result :=
    (Data.MaxBitRate > 0) and (Data.MaxBitRate < 320000) and
    ((Data.Channels = WMA_CM_MONO) or (Data.Channels = WMA_CM_STEREO)) and
    (Data.SampleRate >= 8000) and (Data.SampleRate <= 96000) and
    (Data.ByteRate > 0) and (Data.ByteRate < 40000);
end;

{ --------------------------------------------------------------------------- }

function ExtractTrack2(const TrackString: WideString): Integer;
var
  Value, Code: Integer;
begin
  { Extract track from string }
  Result := 0;
  Val(TrackString, Value, Code);
  if Code = 0 then Result := Value;
end;

{ ********************** Private functions & procedures ********************* }

procedure TWMAfile.FResetData;
begin
  { Reset variables }
  FValid := false;
  FFileSize := 0;
  FChannelModeID := WMA_CM_UNKNOWN;
  FSampleRate := 0;
  FDuration := 0;
  FBitRate := 0;
  FTitle := '';
  FArtist := '';
  FAlbum := '';
  FTrack := 0;
  FYear := '';
  FGenre := '';
  FComment := '';
end;

{ --------------------------------------------------------------------------- }

function TWMAfile.FGetChannelMode: string;
begin
  { Get channel mode name }
  Result := WMA_MODE[FChannelModeID];
end;

{ ********************** Public functions & procedures ********************** }

constructor TWMAfile.Create;
begin
  { Create object }
  inherited;
  FResetData;
end;

{ --------------------------------------------------------------------------- }

function TWMAfile.ReadFromFile(const FileName: string): Boolean;
var
  Data: FileData;
begin
  { Reset variables and load file data }
  FResetData;
  FillChar(Data, SizeOf(Data), 0);
  Result := ReadData(FileName, Data);
  { Process data if loaded and valid }
  if Result and IsValid(Data) then
  begin
    FValid := true;
    { Fill properties with loaded data }
    FFileSize := Data.FileSize;
    FChannelModeID := Data.Channels;
    FSampleRate := Data.SampleRate;
    FDuration := Data.FileSize * 8 / Data.MaxBitRate;
    FBitRate := Data.ByteRate * 8 div 1000;
    FTitle := Trim(Data.Tag[1]);
    FArtist := Trim(Data.Tag[2]);
    FAlbum := Trim(Data.Tag[3]);
    FTrack := ExtractTrack2(Trim(Data.Tag[4]));
    FYear := Trim(Data.Tag[5]);
    FGenre := Trim(Data.Tag[6]);
    FComment := Trim(Data.Tag[7]);
  end;
end;




var
 i : integer; // for genres initialization

initialization
 ID3Genres := TStringList.Create;
 ID3Genres.Duplicates := dupAccept;
 ID3Genres.Sorted := false;
 ID3Genres.CommaText :=
  '"Blues","Classic Rock","Country","Dance","Disco","Funk","Grunge",'
  + '"Hip-Hop","Jazz","Metal","New Age","Oldies","Other","Pop",'
  + '"R&B","Rap","Reggae","Rock","Techno","Industrial","Alternative",'
  + '"Ska","Death Metal","Pranks","Soundtrack","Euro-Techno","Ambient",'
  + '"Trip-Hop","Vocal","Jazz+Funk","Fusion","Trance","Classical",'
  + '"Instrumental","Acid","House","Game","Sound Clip","Gospel",'
  + '"Noise","AlternRock","Bass","Soul","Punk","Space","Meditative",'
  + '"Instrumental Pop","Instrumental Rock","Ethnic","Gothic",'
  + '"Darkwave","Techno-Industrial","Electronic","Pop-Folk","Eurodance",'
  + '"Dream","Southern Rock","Comedy","Cult","Gangsta","Top 40",'
  + '"Christian Rap","Pop/Funk","Jungle","Native American","Cabaret",'
  + '"New Wave","Psychedelic","Rave","Showtunes","Trailer","Lo-Fi",'
  + '"Tribal","Acid Punk","Acid Jazz","Polka","Retro","Musical",'
  + '"Rock & Roll","Hard Rock",' // 79 standard genres
  + '"Folk","Folk/Rock","National Folk","Swing","Fast Fusion",'
  + '"Bebob","Latin","Revival","Celtic","Bluegrass","Avantgarde",'
  + '"Gothic Rock","Progressive Rock","Psychedelic Rock",'
  + '"Symphonic Rock","Slow Rock","Big Band","Chorus","Easy Listening",'
  + '"Acoustic","Humour","Speech","Chanson","Opera","Chamber Music",'
  + '"Sonata","Symphony","Booty Bass","Primus","Porn Groove",'
  + '"Satire","Slow Jam","Club","Tango","Samba","Folklore",' // add
  + '"Ballad","Power Ballad","Rhythmic Soul","Freestyle",'
  + '"Duet","Punk Rock","Drum Solo","Acapella","Euro-House",'
  + '"Dance Hall", "Goa", "Drum & Bass", "Club-House", "Hardcore",'
  + '"Terror", "Indie", "BritPop", "Negerpunk", "Polsk Punk", "Beat",'
  + '"Christian Gangs", "Heavy Metal", "Black Metal", "Crossover",'
  + '"Contemporary Ch?", "Cristian Rock", "Merengue", "Salsa",'
  + '"Thrash Metal", "Anime", "JPop", "Synthpop"'; // 0 .. 93h

 // save genres IDs for genres list resorting ability
 for i := 0 to ID3Genres.Count - 1
  do ID3Genres.Objects [i] := TObject (i);


finalization

 if Assigned (ID3Genres)
  then ID3Genres.Free;

end.

